@@ -1,5 +1,8 @@
+import math
+
 import torch as to
 import torch.nn as nn
+from typing import List
 
 from message_passing_nn.data.data_preprocessor import DataPreprocessor
 from message_passing_nn.model.node import Node
@@ -16,7 +19,7 @@ class GraphEncoder(nn.Module):
         super(GraphEncoder, self).__init__()
         base_4d_tensor_shape = [number_of_nodes, number_of_nodes, number_of_node_features, number_of_node_features]
         base_3d_tensor_shape = [number_of_nodes, number_of_node_features, number_of_node_features]
-        base_2d_tensor_shape = [number_of_node_features]
+        base_2d_tensor_shape = [number_of_node_features, 1]
 
         self.time_steps = time_steps
         self.number_of_nodes = number_of_nodes
@@ -25,17 +28,17 @@ class GraphEncoder(nn.Module):
         self.fully_connected_layer_output_size = fully_connected_layer_output_size
         self.device = device
 
-        self.w_gru_update_gate_features = nn.Parameter(to.randn(base_4d_tensor_shape, device=self.device), requires_grad=True)
-        self.w_gru_forget_gate_features = nn.Parameter(to.randn(base_4d_tensor_shape, device=self.device), requires_grad=True)
-        self.w_gru_current_memory_message_features = nn.Parameter(to.randn(base_4d_tensor_shape, device=self.device), requires_grad=True)
-        self.u_gru_update_gate = nn.Parameter(to.randn(base_4d_tensor_shape, device=self.device), requires_grad=True)
-        self.u_gru_forget_gate = nn.Parameter(to.randn(base_4d_tensor_shape, device=self.device), requires_grad=True)
-        self.u_gru_current_memory_message = nn.Parameter(to.randn(base_4d_tensor_shape, device=self.device), requires_grad=True)
-        self.b_gru_update_gate = nn.Parameter(to.randn(base_2d_tensor_shape, device=self.device), requires_grad=True)
-        self.b_gru_forget_gate = nn.Parameter(to.randn(base_2d_tensor_shape, device=self.device), requires_grad=True)
-        self.b_gru_current_memory_message = nn.Parameter(to.randn(base_2d_tensor_shape, device=self.device), requires_grad=True)
-        self.u_graph_node_features = nn.Parameter(to.randn(base_3d_tensor_shape, device=self.device), requires_grad=True)
-        self.u_graph_neighbor_messages = nn.Parameter(to.randn(base_3d_tensor_shape, device=self.device), requires_grad=True)
+        self.w_gru_update_gate_features = self._get_parameter(base_4d_tensor_shape)
+        self.w_gru_forget_gate_features = self._get_parameter(base_4d_tensor_shape)
+        self.w_gru_current_memory_message_features = self._get_parameter(base_4d_tensor_shape)
+        self.u_gru_update_gate = self._get_parameter(base_4d_tensor_shape)
+        self.u_gru_forget_gate = self._get_parameter(base_4d_tensor_shape)
+        self.u_gru_current_memory_message = self._get_parameter(base_4d_tensor_shape)
+        self.b_gru_update_gate = self._get_parameter(base_2d_tensor_shape).view(base_2d_tensor_shape[0])
+        self.b_gru_forget_gate = self._get_parameter(base_2d_tensor_shape).view(base_2d_tensor_shape[0])
+        self.b_gru_current_memory_message = self._get_parameter(base_2d_tensor_shape).view(base_2d_tensor_shape[0])
+        self.u_graph_node_features = self._get_parameter(base_3d_tensor_shape)
+        self.u_graph_neighbor_messages = self._get_parameter(base_3d_tensor_shape)
         self.linear = to.nn.Linear(self.fully_connected_layer_input_size, self.fully_connected_layer_output_size)
         self.sigmoid = to.nn.Sigmoid()
 
@@ -73,7 +76,8 @@ class GraphEncoder(nn.Module):
         return encodings
 
     def _send_messages(self, node_features: to.Tensor, adjacency_matrix: to.Tensor) -> to.Tensor:
-        messages = to.zeros((self.number_of_nodes, self.number_of_nodes, self.number_of_node_features), device=self.device)
+        messages = to.zeros((self.number_of_nodes, self.number_of_nodes, self.number_of_node_features),
+                            device=self.device)
         for step in range(self.time_steps):
             messages = self._compose_messages(node_features, adjacency_matrix, messages)
         return messages
@@ -114,12 +118,12 @@ class GraphEncoder(nn.Module):
         update_gate = self._pass_through_update_gate(messages, node, end_node_id, node_features)
         current_memory = self._get_current_memory_message(messages, node, end_node_id, node_features)
         return to.add(
-                    to.mul(
-                        to.sub(to.ones(update_gate.shape, device=self.device),
-                               update_gate),
-                        previous_messages),
-                    to.mul(update_gate,
-                           current_memory))
+            to.mul(
+                to.sub(to.ones(update_gate.shape, device=self.device),
+                       update_gate),
+                previous_messages),
+            to.mul(update_gate,
+                   current_memory))
 
     def _sum_messages_from_neighbors_except_target(self,
                                                    messages: to.Tensor,
@@ -143,7 +147,8 @@ class GraphEncoder(nn.Module):
         update_gate_output = to.sigmoid(
             to.add(
                 to.add(self.w_gru_update_gate_features[node.node_id, end_node_id].matmul(node_features[node.node_id]),
-                       self.u_gru_update_gate[node.node_id, end_node_id].matmul(message_from_a_neighbor_other_than_target)),
+                       self.u_gru_update_gate[node.node_id, end_node_id].matmul(
+                           message_from_a_neighbor_other_than_target)),
                 self.b_gru_update_gate))
         return update_gate_output
 
@@ -157,8 +162,10 @@ class GraphEncoder(nn.Module):
                                                                           end_node_id,
                                                                           node_features)
         current_memory_message = to.add(
-            to.add(self.w_gru_current_memory_message_features[node.node_id, end_node_id].matmul(node_features[node.node_id]),
-                   self.u_gru_current_memory_message[node.node_id, end_node_id].matmul(messages_passed_through_reset_gate)),
+            to.add(self.w_gru_current_memory_message_features[node.node_id, end_node_id].matmul(
+                node_features[node.node_id]),
+                   self.u_gru_current_memory_message[node.node_id, end_node_id].matmul(
+                       messages_passed_through_reset_gate)),
             self.b_gru_current_memory_message)
         return to.tanh(current_memory_message)
 
@@ -170,9 +177,12 @@ class GraphEncoder(nn.Module):
         return self.u_gru_current_memory_message[node.node_id, end_node_id].matmul(sum([to.mul(to.sigmoid(
             to.add(
                 to.add(self.w_gru_update_gate_features[node.node_id, end_node_id].matmul(node_features[node.node_id]),
-                       self.u_gru_update_gate[node.node_id, end_node_id].matmul(messages[node.node_id, reset_node_index])),
+                       self.u_gru_update_gate[node.node_id, end_node_id].matmul(
+                           messages[node.node_id, reset_node_index])),
                 self.b_gru_update_gate)).long(), messages[node.node_id, reset_node_index])
-                                                                         for reset_node_index in node.get_start_node_neighbors_without_end_node(end_node_id)[0]]))
+                                                                                        for reset_node_index in
+                                                                                        node.get_start_node_neighbors_without_end_node(
+                                                                                            end_node_id)[0]]))
 
     def _pass_through_reset_gate(self,
                                  messages: to.Tensor,
@@ -185,6 +195,9 @@ class GraphEncoder(nn.Module):
                        self.u_gru_update_gate[node.node_id, end_node_id].matmul(messages[node.node_id, end_node_id])),
                 self.b_gru_update_gate)).long()
 
+    def _get_parameter(self, tensor_shape: List[int]) -> nn.Parameter:
+        return nn.Parameter(nn.init.kaiming_normal_(to.zeros(tensor_shape, device=self.device)), requires_grad=True)
+
     @staticmethod
     def _create_node(node_features: to.Tensor, adjacency_matrix: to.Tensor, node_id: int) -> Node:
         return Node(node_features, adjacency_matrix, node_id)
